#!/usr/bin/env python3
"""
Writing to a dump file.

This file is part of pg_seldump.
"""

import math
import logging
from datetime import datetime

import psycopg
from psycopg import sql

from .consts import PROJECT_URL, VERSION
from .writer import Writer
from .exceptions import DumpError

logger = logging.getLogger("seldump.dumpwriter")


class DumpWriter(Writer):
    def __init__(self, outfile, reader):
        self.outfile = outfile
        self.reader = reader

        self._start_time = None
        self._copy_start_pos = None
        self._copy_size = None

    def dump_table(self, table, match):
        logger.info("writing %s %s", table.kind, table)

        self._begin_table(table)
        self._copy_table(table, match)
        self._end_table(table)

    def _begin_table(self, table):
        self.write(
            sql.SQL("\nalter table {} disable trigger all;\n").format(table.ident)
        )

    def _end_table(self, table):
        self.write(
            sql.SQL("\nalter table {} enable trigger all;\n\n").format(table.ident)
        )

        if self._copy_size is not None:
            if self._copy_size >= 1024:
                pretty = " (%s)" % pretty_size(self._copy_size)
            else:
                pretty = ""

            self.write(
                "-- %s bytes written for table %s%s\n\n"
                % (self._copy_size, table, pretty)
            )

    def _copy_table(self, table, match):
        assert match.import_statement
        assert match.copy_statement

        self.write(match.import_statement)

        stmt = self.reader.obj_as_string(match.copy_statement)
        logger.debug("exporting using: %s", stmt)

        self._begin_copy()
        try:
            self.reader.copy(stmt, self.outfile)
        except psycopg.DatabaseError as e:
            logger.error("failed copy using statement:\n\n%s\n", stmt)
            raise DumpError("failed to copy from table %s: %s" % (table, e))
        else:
            self.write("\\.\n")

        self._end_copy()

    def dump_sequence(self, seq, match):
        logger.info("writing %s %s", seq.kind, seq)

        # Escape the sequence as identifier then as string to make a value
        # good for regclass
        name = sql.Literal(self.reader.obj_as_string(seq.ident))
        val = sql.Literal(self.reader.get_sequence_value(seq))
        stmt = sql.SQL("\nselect pg_catalog.setval({}, {}, true);\n\n").format(
            name, val
        )
        self.write(stmt)

    def dump_materialized_view(self, matview, match):
        logger.info("writing %s %s", matview.kind, matview)

        self.write(sql.SQL("\nrefresh materialized view {};\n").format(matview.ident))

    def begin_dump(self):
        self.write("-- PostgreSQL data dump generated by pg_seldump %s\n" % VERSION)
        self.write("-- %s\n\n" % PROJECT_URL)

        self._start_time = now = datetime.utcnow()
        self.write("-- Data dump started at %sZ\n\n" % now)

        self.write("set session authorization default;\n")

    def end_dump(self):
        self.write("\n\nanalyze;\n\n")

        now = datetime.utcnow()
        elapsed = pretty_timedelta(now - self._start_time)
        self.write("-- Data dump finished at %sZ (%s)\n\n" % (now, elapsed))

        # No highlight please
        self.write("-- vim: set filetype=:\n")

    def write(self, data):
        if isinstance(data, sql.Composable):
            data = self.reader.obj_as_string(data)

        if isinstance(data, str):
            data = data.encode("utf8")

        self.outfile.write(data)

    def _begin_copy(self):
        """
        Mark the start of the copy of a table data.

        Memorize where we are in the file output file, if the file is seekable.
        """
        if self.outfile.seekable():
            self._copy_start_pos = self.outfile.tell()

    def _end_copy(self):
        """
        Mark the end of the copy of a table data.

        If the file is seekable return the amout of bytes copied.
        """
        if self.outfile.seekable() and self._copy_start_pos is not None:
            self._copy_size = self.outfile.tell() - self._copy_start_pos
            self._copy_start_pos = None


def pretty_size(size):
    """
    Display a size in bytes in a human friendly way
    """
    if size <= 0:
        # Not bothering with negative numbers
        return "%sB" % size

    suffixes = ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
    i = int(math.floor(math.log(size, 1024)))
    p = math.pow(1024, i)
    s = round(size / p, 2)
    return "%s %s" % (s, suffixes[i])


def pretty_timedelta(delta):
    """
    Display a time interval in a human friendly way
    """
    rem, secs = divmod(abs(delta.total_seconds()), 60)
    rem, mins = divmod(rem, 60)
    days, hours = divmod(rem, 24)
    parts = [(days, "d"), (hours, "h"), (mins, "m"), (secs, "s")]
    while parts and parts[0][0] == 0:
        del parts[0]
    sign = "-" if delta.total_seconds() < 0 else ""
    return sign + " ".join("%.0f%s" % p for p in parts)



"""
Tests that create schema and data in a real db, dump the data, wipe the db,
restore the data, and verify that stuff is what expected.
"""


def test_a_simple_thing(db, dbdumper, psql):
    """
    With the tables::

        table1:dump -> table2:ref -> table3:ref
        [data <= 'c']

        table4:ignored

    Only required data in tables 1, 2, 3 is dumped.
    Sequences 1, 2, 3 are dumped, seq/table 4 skipped.
    """
    objs = db.create_sample(
        4,
        fkeys=[
            ("table1", "t2id", "table2", "id"),
            ("table2", "t3id", "table3", "id"),
        ],
    )
    db.write_schema(objs)
    db.fill_data("table3", "data", "ijkl")
    db.fill_data("table2", ("data", "t3id"), "efgh", (1, 1, 2, 3))
    db.fill_data("table1", ("data", "t2id"), "abcd", (1, 1, 2, 3))
    db.fill_data("table4", "data", "ijkl")

    dbdumper.reader.load_schema()
    dbdumper.add_config(
        """
db_objects:
- name: table1
  filter: data <= 'c'
"""
    )
    dbdumper.perform_dump()

    db.truncate(dbdumper.db)
    psql.load_file(dbdumper.writer.outfile)

    conn = dbdumper.reader.connection
    with conn.transaction():
        with conn.cursor() as cur:
            cur.execute("select id, data, t2id from table1 order by id")
            assert cur.fetchall() == [(1, "a", 1), (2, "b", 1), (3, "c", 2)]
            cur.execute("select id, data, t3id from table2 order by id")
            assert cur.fetchall() == [(1, "e", 1), (2, "f", 1)]
            cur.execute("select id, data from table3 order by id")
            assert cur.fetchall() == [(1, "i")]
            cur.execute("select id, data from table4 order by id")
            assert cur.fetchall() == []
            cur.execute("select nextval('table1_id_seq')")
            assert cur.fetchone()[0] == 5
            cur.execute("select nextval('table3_id_seq')")
            assert cur.fetchone()[0] == 5
            cur.execute("select nextval('table4_id_seq')")
            assert cur.fetchone()[0] == 1


def test_columns_edit(db, dbdumper, psql):
    """
    Some columns can be omitted, some can be changed.
    """

    objs = db.create_sample(
        3,
        fkeys=[
            ("table1", "t2id", "table2", "id"),
            ("table2", "t3id", "table3", "id"),
        ],
    )

    for i in range(2):
        objs[i]["columns"]["password"] = {"name": "password", "type": "text"}

    db.write_schema(objs)
    db.fill_data("table2", ("data", "password"), "efgh", ["pass"] * 4)
    db.fill_data(
        "table1", ("data", "t2id", "password"), "abcd", (1, 1, 2, 3), ["pass"] * 4
    )

    dbdumper.reader.load_schema()
    dbdumper.add_config(
        """
db_objects:
- name: table1
  filter: data <= 'c'
  no_columns:
  - password
  replace:
    data: "'x'"

- name: table2
  action: ref
  no_columns:
  - password
  replace:
    data: "'y'"
"""
    )
    dbdumper.perform_dump()

    db.truncate(dbdumper.db)
    psql.load_file(dbdumper.writer.outfile)

    conn = dbdumper.reader.connection
    with conn.transaction():
        with conn.cursor() as cur:
            cur.execute("select id, data, password, t2id from table1 order by id")
            assert cur.fetchall() == [
                (1, "x", None, 1),
                (2, "x", None, 1),
                (3, "x", None, 2),
            ]
            cur.execute("select id, data, password from table2 order by id")
            assert cur.fetchall() == [(1, "y", None), (2, "y", None)]


def test_two_dependencies(db, dbdumper, psql):
    """
    A fkey table can be referred by more than one table.

        table1:dump ->
                        table3:ref -> table4:ref
        table2:dump ->
    """

    objs = db.create_sample(
        4,
        fkeys=[
            ("table1", "t13id", "table3", "id"),
            ("table2", "t23id", "table3", "id"),
            ("table3", "t4id", "table4", "id"),
        ],
    )

    db.write_schema(objs)
    db.fill_data("table4", ("data"), "opqrst")
    db.fill_data("table3", ("data", "t4id"), "ijklmn", reversed(range(1, 7)))
    db.fill_data("table1", ("data", "t13id"), "abcd", (1, 2, 3, 4))
    db.fill_data("table2", ("data", "t23id"), "efgh", (2, 3, 4, 5))

    dbdumper.reader.load_schema()
    dbdumper.add_config(
        """
db_objects:
- name: table1
  filter: data <= 'b'

- name: table2
  filter: data <= 'f'
"""
    )
    dbdumper.perform_dump()

    db.truncate(dbdumper.db)
    psql.load_file(dbdumper.writer.outfile)

    conn = dbdumper.reader.connection
    with conn.transaction():
        with conn.cursor() as cur:
            cur.execute("select id, data, t13id from table1 order by id")
            assert cur.fetchall() == [(1, "a", 1), (2, "b", 2)]
            cur.execute("select id, data, t23id from table2 order by id")
            assert cur.fetchall() == [(1, "e", 2), (2, "f", 3)]
            cur.execute("select id, data, t4id from table3 order by id")
            assert cur.fetchall() == [(1, "i", 6), (2, "j", 5), (3, "k", 4)]
            cur.execute("select id, data from table4 order by id")
            assert cur.fetchall() == [(4, "r"), (5, "s"), (6, "t")]


def test_dump_and_ref(db, dbdumper, psql):
    """
    Tables can have some records referred, some dumped.

    table1:sel -> table2:sel ->table3
    [where1]      [where2]
    """
    objs = db.create_sample(
        3,
        fkeys=[
            ("table1", "t2id", "table2", "id"),
            ("table2", "t3id", "table3", "id"),
        ],
    )
    db.write_schema(objs)
    db.fill_data("table3", "data", "ijkl")
    db.fill_data("table2", ("data", "t3id"), "efgh", (1, 2, 3, 4))
    db.fill_data("table1", ("data", "t2id"), "abcd", (1, 2, 3, 4))

    dbdumper.reader.load_schema()
    dbdumper.add_config(
        """
db_objects:
- name: table1
  filter: data = 'a'
- name: table2
  filter: data = 'g'
"""
    )
    dbdumper.perform_dump()

    db.truncate(dbdumper.db)
    psql.load_file(dbdumper.writer.outfile)

    conn = dbdumper.reader.connection
    with conn.transaction():
        with conn.cursor() as cur:
            cur.execute("select id, data, t2id from table1 order by id")
            assert cur.fetchall() == [(1, "a", 1)]
            cur.execute("select id, data, t3id from table2 order by id")
            assert cur.fetchall() == [(1, "e", 1), (3, "g", 3)]
            cur.execute("select id, data from table3 order by id")
            assert cur.fetchall() == [(1, "i"), (3, "k")]


def test_self_ref(db, dbdumper, psql):
    """
    Self-referring tables can get other records dumped.
    """
    objs = db.create_sample(
        1,
        fkeys=[
            ("table1", "father_id", "table1", "id"),
            ("table1", "mother_id", "table1", "id"),
        ],
    )
    db.write_schema(objs)
    db.fill_data(
        "table1",
        ("data", "father_id", "mother_id"),
        "abcdef",
        [None, None, 1, None, 4, None],
        [None, None, 2, None, 3, 3],
    )

    dbdumper.reader.load_schema()
    dbdumper.add_config(
        """
db_objects:
- name: table1
  filter: data = 'e'
"""
    )
    dbdumper.perform_dump()

    db.truncate(dbdumper.db)
    psql.load_file(dbdumper.writer.outfile)

    conn = dbdumper.reader.connection
    with conn.transaction():
        with conn.cursor() as cur:
            cur.execute("select id, data, father_id, mother_id from table1 order by id")
            assert cur.fetchall() == [
                (1, "a", None, None),
                (2, "b", None, None),
                (3, "c", 1, 2),
                (4, "d", None, None),
                (5, "e", 4, 3),
            ]


def test_ref_to_self_ref(db, dbdumper, psql):
    """
    Records from a self referring table and external one are dumped

    table1 -> table2 -+ -> table3
                 ^    |
                 +----+
    """
    objs = db.create_sample(
        3,
        fkeys=[
            ("table1", "t2id", "table2", "id"),
            ("table2", "father_id", "table2", "id"),
            ("table2", "mother_id", "table2", "id"),
            ("table2", "t3id", "table3", "id"),
        ],
    )
    db.write_schema(objs)
    db.fill_data("table3", ("data",), "stuvwxyz")
    db.fill_data(
        "table2",
        ("data", "father_id", "mother_id", "t3id"),
        "abcdefgh",
        [None, None, 1, None, 4, None, None, 6],
        [None, None, 2, None, 3, 3, None, None],
        range(1, 9),
    )
    db.fill_data("table1", ("data", "t2id"), "ijk", (5, 6, 7))

    dbdumper.reader.load_schema()
    dbdumper.add_config(
        """
db_objects:
- name: table1
  filter: data = any('{i,k}')
- name: table2
  action: ref
  no_columns: []
"""
    )
    dbdumper.perform_dump()

    db.truncate(dbdumper.db)
    psql.load_file(dbdumper.writer.outfile)

    conn = dbdumper.reader.connection
    with conn.transaction():
        with conn.cursor() as cur:
            cur.execute("select id, data, father_id, mother_id from table2 order by id")
            assert cur.fetchall() == [
                (1, "a", None, None),
                (2, "b", None, None),
                (3, "c", 1, 2),
                (4, "d", None, None),
                (5, "e", 4, 3),
                (7, "g", None, None),
            ]

            cur.execute("select id, data from table3 order by id")
            want = [(1, "s"), (2, "t"), (3, "u"), (4, "v"), (5, "w"), (7, "y")]
            assert cur.fetchall() == want



"""
Program constants.

This file is part of pg_seldump.
"""

VERSION = "0.3.1"
PROJECT_URL = "https://github.com/dvarrazzo/pg_seldump"


# kinds with a state to be dumped
# other objects are either internal (indexes, toast tables) or stateless (views)
DUMPABLE_KINDS = set("rSmp")

KIND_TABLE = "table"
KIND_MATVIEW = "materialized view"
KIND_SEQUENCE = "sequence"
KIND_PART_TABLE = "partitioned table"

# relkind values: https://www.postgresql.org/docs/11/catalog-pg-class.html
PG_KINDS = {
    "r": KIND_TABLE,
    "i": "index",
    "S": KIND_SEQUENCE,
    "t": "toast table`",
    "v": "view",
    "m": KIND_MATVIEW,
    "c": "composite type",
    "f": "foreign table",
    "p": KIND_PART_TABLE,
    "I": "partitioned index",
}

# reverse map from kind name to relkind
REVKINDS = {v: k for k, v in PG_KINDS.items()}



#!/usr/bin/env python3
"""
Create a selective dump of a PostgreSQL database.
"""

# This file is part of pg_seldump.

import sys
import logging
from signal import SIGPIPE
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from .config import load_config
from .consts import VERSION
from .dumper import Dumper
from .dbreader import DbReader
from .dumpwriter import DumpWriter
from .dummywriter import DummyWriter
from .exceptions import SelDumpException, ConfigError

logger = logging.getLogger("seldump")


def main():
    """Run the program, raise exceptions."""
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s"
    )
    opt = parse_cmdline()
    logger.setLevel(opt.loglevel)

    writer = DummyWriter()
    reader = DbReader(opt.dsn)
    dumper = Dumper(reader=reader, writer=writer)

    # Parse all the config files and look for all the errors.
    # Bail out if there is any error
    confs = [load_config(fn) for fn in opt.config_files]
    if [conf for conf in confs if conf is None]:
        return 1

    for conf in confs:
        dumper.add_config(conf)

    # Plan now. before opening the output file to write.
    reader.load_schema()
    dumper.plan_dump()

    should_close = False
    if not opt.test:
        if opt.outfile != "-":
            try:
                outfile = open(opt.outfile, "wb")
                should_close = True
            except Exception as e:
                raise ConfigError("couldn't open %s for writing: %s" % (outfile, e))
        else:
            outfile = sys.stdout.buffer

        writer = DumpWriter(outfile=outfile, reader=reader)
        dumper.writer = writer

    try:
        dumper.run_dump()
    finally:
        if should_close:
            outfile.close()


def script():
    """Run the program and terminate the process."""
    try:
        sys.exit(main())

    except SelDumpException as e:
        if str(e):
            logger.error("%s", e)
        sys.exit(1)

    except BrokenPipeError as e:
        logger.error("dump interrupted: %s", e)
        # Not entirely correct: might have been ESHUTDOWN
        sys.exit(SIGPIPE + 128)

    except Exception:
        logger.exception("unexpected error")
        sys.exit(1)

    except KeyboardInterrupt:
        logger.info("user interrupt")
        sys.exit(1)


def parse_cmdline():
    parser = ArgumentParser(
        description=__doc__, formatter_class=RawDescriptionHelpFormatter
    )

    parser.add_argument("--version", action="version", version="%%(prog)s %s" % VERSION)

    parser.add_argument(
        "config_files",
        nargs="+",
        metavar="config",
        help="yaml file describing the data to dump",
    )

    parser.add_argument(
        "--dsn",
        default="",
        help="database connection string [default: %(default)r]",
    )

    parser.add_argument(
        "--outfile",
        "-o",
        default="-",
        help="the file where to save the dump [default: stdout]",
    )

    parser.add_argument(
        "--test",
        action="store_true",
        help="test the configuration to verify it works as expected",
    )

    g = parser.add_mutually_exclusive_group()
    g.add_argument(
        "-q",
        "--quiet",
        help="talk less",
        dest="loglevel",
        action="store_const",
        const=logging.WARN,
        default=logging.INFO,
    )
    g.add_argument(
        "-v",
        "--verbose",
        help="talk more",
        dest="loglevel",
        action="store_const",
        const=logging.DEBUG,
        default=logging.INFO,
    )

    opt = parser.parse_args()

    return opt



import pytest

from seldump.dbobjects import Table, Sequence
from seldump.exceptions import ConfigError


def test_void(dumper):
    """On empty input, result is empty."""
    dumper.perform_dump()
    assert not dumper.writer.dumped


def test_one_table(dumper, db):
    """You can select a single table to dump."""
    dumper.reader.load_db(db.create_sample(2))
    dumper.add_config({"db_objects": [{"name": "table1"}]})
    dumper.perform_dump()

    objs = [obj for obj, rule in dumper.writer.dumped]

    obj = dumper.db.get("public", "table1")
    assert isinstance(obj, Table)
    assert obj in objs

    obj = dumper.db.get("public", "table2")
    assert isinstance(obj, Table)
    assert obj not in objs

    obj = dumper.db.get("public", "table1_id_seq")
    assert isinstance(obj, Sequence)
    assert obj in objs

    obj = dumper.db.get("public", "table2_id_seq")
    assert isinstance(obj, Sequence)
    assert obj not in objs


@pytest.mark.parametrize("spec", ["^table[13]$", ["table1", "table3"]])
def test_names(dumper, db, spec):
    """Tables can be selected by names regexp or list"""
    dumper.reader.load_db(db.create_sample(3))
    dumper.add_config({"db_objects": [{"names": spec}]})
    dumper.perform_dump()

    tables = [obj for obj, rule in dumper.writer.dumped if isinstance(obj, Table)]
    assert len(tables) == 2
    assert dumper.db.get("public", "table1") in tables
    assert dumper.db.get("public", "table2") not in tables
    assert dumper.db.get("public", "table3") in tables


def test_name_over_names_regexp(dumper, db):
    """A specified name overrides a names regexp"""
    dumper.reader.load_db(db.create_sample(3))
    dumper.add_config(
        {
            "db_objects": [
                {"name": "table2", "action": "skip"},
                {"names": "^table.$"},
            ]
        }
    )
    dumper.perform_dump()

    tables = [obj for obj, rule in dumper.writer.dumped if isinstance(obj, Table)]
    assert len(tables) == 2
    assert dumper.db.get("public", "table1") in tables
    assert dumper.db.get("public", "table2") not in tables
    assert dumper.db.get("public", "table3") in tables


@pytest.mark.parametrize("bias", [-1, 0, 1])
def test_name_same_as_names_list(dumper, db, bias):
    """List of names have the same precedence over a single name"""
    dumper.reader.load_db(db.create_sample(2))
    dumper.add_config(
        {
            "db_objects": [
                {"name": "table2", "action": "skip", "adjust_score": bias},
                {"names": ["table1", "table2"]},
            ]
        }
    )
    if bias == 0:
        with pytest.raises(ConfigError):
            dumper.perform_dump()
        return

    dumper.perform_dump()

    tables = [obj for obj, rule in dumper.writer.dumped if isinstance(obj, Table)]
    assert dumper.db.get("public", "table1") in tables

    if bias < 0:
        assert len(tables) == 2
        assert dumper.db.get("public", "table2") in tables
    else:
        assert len(tables) == 1
        assert dumper.db.get("public", "table2") not in tables



import re
import logging
import pkg_resources

import yaml
from jsonschema import Draft7Validator

from .yaml import load_yaml

logger = logging.getLogger("seldump.config")

validator = Draft7Validator(
    schema=yaml.load(
        pkg_resources.resource_string("seldump", "schema/config.yaml"),
        Loader=yaml.SafeLoader,
    )
)


def load_config(filename):
    """
    Load and validate a configuration file.

    Return the content as a Python object, validated according to the
    ``config.yaml`` schema, else None (and log about errors).
    """
    try:
        with open(filename) as f:
            conf = load_yaml(f)
    except Exception as e:
        logger.error("loading %s: %s", filename, e)
        return None

    errors = get_config_errors(conf, filename)
    if errors:
        for error in errors:
            logger.error("%s", error)
        return None

    return conf


def get_config_errors(conf, filename="<no name>"):
    """
    Validate a configuration object and return the list of errors found.
    """
    rv = []

    # Give a clearer error message than what jsonschema would give
    # Something like: None is not of type 'object'
    if not isinstance(conf, dict):
        msg = "config must be an object containing 'db_objects'"
        rv.append(located_message(None, filename, msg))
        return rv

    errors = list(validator.iter_errors(conf))
    for error in errors:
        loc = location_from_error(conf, error)
        rv.append(located_message(loc, filename, error.message))

    if isinstance(conf.get("db_objects"), list):
        for obj in conf["db_objects"]:
            if isinstance(obj, dict):
                rv.extend(_get_rule_errors(obj, filename))

    # sort by line number
    def lineno(s):
        m = re.search(r":(\d+)", s)
        return int(m.group(1)) if m is not None else 0

    rv.sort(key=lineno)

    return rv


def _get_rule_errors(obj, filename):
    """
    Return additional errors on a configuration objects.

    I don't see obvious ways to return these errors from jsonschema validation.
    """
    rv = []
    if "name" in obj and "names" in obj:
        loc = location_from_attribs(obj, "name", "names")
        msg = "can't specify both 'name' and 'names'"
        rv.append(located_message(loc, filename, msg))

    if isinstance(obj.get("names"), str):
        try:
            re.compile(obj["names"], re.VERBOSE)
        except re.error as e:
            msg = "names: not a valid regular expression: %s" % e
            loc = location_from_attribs(obj, "names")
            rv.append(located_message(loc, filename, msg))

    if "schema" in obj and "schemas" in obj:
        loc = location_from_attribs(obj, "schema", "schemas")
        msg = "can't specify both 'schema' and 'schemas'"
        rv.append(located_message(loc, filename, msg))

    if isinstance(obj.get("schemas"), str):
        try:
            re.compile(obj["schemas"], re.VERBOSE)
        except re.error as e:
            msg = "schemas: not a valid regular expression: %s" % e
            loc = location_from_attribs(obj, "schemas")
            rv.append(located_message(loc, filename, msg))

    if "kind" in obj and "kinds" in obj:
        loc = location_from_attribs(obj, "kind", "kinds")
        msg = "can't specify both 'kind' and 'kinds'"
        rv.append(located_message(loc, filename, msg))

    return rv


def located_message(loc, filename, message):
    """
    Add location informations to a message string.
    """
    if loc:
        return "at %s: %s" % (loc, message)
    else:
        return "in %s: %s" % (filename, message)


def location_from_attribs(conf, *items):
    """
    Return location information from an attrib on a dict parsed from yaml.
    """
    assert items
    filename = getattr(conf, "filename", None)
    itemlines = getattr(conf, "itemlines", None)
    if not (filename and itemlines):
        return
    try:
        linenos = [itemlines[item] for item in items]
    except (KeyError, IndexError):
        return

    return "%s:%s" % (filename, max(linenos))


def location_from_error(conf, error):
    """
    Return location information from a yaml validation error.
    """
    if error.validator == "additionalProperties":
        rv = _location_from_addprops(conf, error)
        if rv is not None:
            return rv

    # find the closest location for the error
    trail = [conf]
    for item in error.path:
        trail.append(trail[-1][item])

    # Does the last element in the error trail have a position?
    filename = getattr(trail[-1], "filename", None)
    lineno = getattr(trail[-1], "lineno", None)
    if filename and lineno:
        return "%s:%s" % (filename, lineno)

    if len(trail) < 2:
        return

    # Is the last element an item into a container that knows positions?
    filename = getattr(trail[-2], "filename", None)
    itemlines = getattr(trail[-2], "itemlines", None)
    if filename and itemlines:
        try:
            lineno = itemlines[error.path[-1]]
        except (KeyError, IndexError):
            # bah, I give up
            return
        else:
            rv = "%s:%s" % (filename, lineno)
            if isinstance(trail[-2], dict):
                # also add the attribute name
                rv = "%s: %s" % (rv, error.path[-1])
            return rv


def _location_from_addprops(conf, error):
    # Special-case this error otherwise it will report the object position
    # not the attribute one.
    filename = getattr(error.instance, "filename", None)
    itemlines = getattr(error.instance, "itemlines", None)
    if not (filename and itemlines):
        return

    # parse the attr name from the error message, I don't see it elsewhere
    m = re.search(r"'([^']*)' was unexpected", error.message)
    if m is None:
        return

    attr = m.group(1)
    if attr in itemlines:
        return "%s:%s" % (filename, itemlines[attr])



#!/usr/bin/env python3
"""
Objects to generate dynamic queries.

These objects create a small and specific ORM to help generating queries
dynamically.

You can create queries by nesting `QueryNode` subclasses nodes, and use
`SqlQueryVisitor.visit()` to return the SQL statement as
`psycopg.sql.Composable` object.

This file is part of pg_seldump.
"""

from psycopg import sql

from .nodes import NodeVisitor
from .dbobjects import Table


class QueryNode:
    def as_string(self):
        return PrintQueryVisitor().as_string(self)


class Query(QueryNode):
    pass


class CopyOut(QueryNode):
    def __init__(self, source):
        self.source = source


class Select(Query):
    def __init__(self, columns, from_, where=None):
        self.columns = columns
        self.from_ = from_
        self.where = where


class RecursiveCTE(Query):
    def __init__(self, name, base_query, rec_cond):
        self.name = name
        self.base_query = base_query
        self.rec_cond = rec_cond


class Union(Query):
    def __init__(self, queries):
        self.queries = queries


class FromEntry(QueryNode):
    def __init__(self, source, alias=None):
        self.source = source
        self.alias = alias


class Predicate(QueryNode):
    pass


class Exists(Predicate):
    def __init__(self, query):
        self.query = query


class Or(Predicate):
    def __init__(self, conds):
        self.conds = conds


class And(Predicate):
    def __init__(self, conds):
        self.conds = conds


class FkeyJoin(Predicate):
    def __init__(self, fkey, from_, to):
        self.fkey = fkey
        self.from_ = from_
        self.to = to


class SqlQueryVisitor(NodeVisitor):
    def __init__(self):
        self._level = 0
        self._first = True

    def indent(self):
        self._level += 4
        self._first = False

    def dedent(self):
        self._level -= 4
        assert self._level >= 0

    def indented(self, obj):
        if self._level or not self._first:
            return sql.Composed([sql.SQL("\n"), sql.SQL(" " * self._level), obj])
        else:
            return obj

    def visit_CopyOut(self, copy):
        parts = []
        parts.append(self.indented(sql.SQL("copy (")))
        self.indent()
        parts.append(self.visit(copy.source))
        self.dedent()
        parts.append(self.indented(sql.SQL(") to stdout")))
        return sql.SQL(" ").join(parts)

    def visit_Select(self, select):
        parts = []
        if isinstance(select.from_.source, RecursiveCTE):
            parts.append(self.indented(sql.SQL("with")))
            parts.append(self.visit(select.from_))

        parts.append(self.indented(sql.SQL("select")))
        parts.append(self._cols_list(select))
        parts.append(self.indented(sql.SQL("from")))

        if not isinstance(select.from_.source, RecursiveCTE):
            parts.append(self.visit(select.from_))
        else:
            parts.append(sql.Identifier(select.from_.source.name))

        if select.where:
            parts.append(self.indented(sql.SQL("where")))
            parts.append(self.visit(select.where))

        return sql.SQL(" ").join(parts)

    def visit_RecursiveCTE(self, cte):
        parts = []
        parts.append(sql.SQL("recursive"))
        parts.append(sql.Identifier(cte.name))
        parts.append(sql.SQL("as ("))
        self.indent()

        parts.append(self.visit(cte.base_query))

        parts.append(self.indented(sql.SQL("union")))

        parts.append(self.indented(sql.SQL("select")))
        parts.append(self._cols_list(cte.base_query))
        parts.append(self.indented(sql.SQL("from")))
        assert cte.base_query.from_.alias
        parts.append(self.visit(cte.base_query.from_))
        parts.append(self.indented(sql.SQL("join")))
        parts.append(sql.Identifier(cte.name))
        parts.append(sql.SQL("on"))
        parts.append(self.visit(cte.rec_cond))

        self.dedent()
        parts.append(self.indented(sql.SQL(")")))

        return sql.SQL(" ").join(parts)

    def visit_FromEntry(self, from_):
        if isinstance(from_.source, sql.Identifier):
            rv = sql.Composed([sql.SQL("only "), from_.source])
        elif isinstance(from_.source, Table):
            rv = sql.Composed([sql.SQL("only "), from_.source.ident])
        elif isinstance(from_.source, QueryNode):
            rv = self.visit(from_.source)
        else:
            raise TypeError("can't deal with %s in a 'from'", from_.source)

        if from_.alias and not isinstance(from_.source, RecursiveCTE):
            rv = sql.SQL("{} as {}").format(rv, sql.Identifier(from_.alias))

        return rv

    def visit_Exists(self, exists):
        rv = []
        rv.append(sql.SQL("exists ("))
        self.indent()
        rv.append(self.visit(exists.query))
        self.dedent()
        rv.append(sql.SQL(")"))
        return sql.Composed(rv)

    def visit_And(self, node, kw="and"):
        rv = [sql.SQL("(")]
        self.indent()
        for i, cond in enumerate(node.conds):
            if i:
                rv.append(self.indented(sql.SQL(kw + " ")))
                rv.append(self.visit(cond))
            else:
                rv.append(self.indented(self.visit(cond)))
        rv.append(sql.SQL(")"))
        self.dedent()
        return sql.Composed(rv)

    def visit_Or(self, node):
        return self.visit_And(node, kw="or")

    def visit_FkeyJoin(self, join):
        if len(join.fkey.table_cols) == 1:
            lhs = sql.Identifier(join.from_, join.fkey.table_cols[0])
            rhs = sql.Identifier(join.to, join.fkey.ftable_cols[0])
        else:
            lhs = sql.SQL(", ").join(
                sql.Identifier(join.from_, col) for col in join.fkey.table_cols
            )
            rhs = sql.SQL(", ").join(
                sql.Identifier(join.to, col) for col in join.fkey.ftable_cols
            )

        return sql.SQL("(({}) = ({}))").format(lhs, rhs)

    def visit_Composable(self, obj):
        return obj

    def _cols_list(self, select):
        cols = []
        for col in select.columns:
            if isinstance(col, sql.Composable):
                cols.append(col)
            elif isinstance(col, str):
                cols.append(sql.Identifier(col))
            else:
                raise TypeError("bad column: %s" % col)

        return sql.SQL(", ").join(cols)


class PrintQueryVisitor(NodeVisitor):
    def __init__(self):
        self.reset()

    def reset(self):
        self._stack = []
        self.output = []
        self._level = 0

    def as_string(self, node):
        self.reset()
        self.visit(node)
        return "\n".join(self.output)

    def visit_QueryNode(self, node):
        line = [node.__class__.__name__]
        if not self.empty():
            line.insert(0, "%s:" % self.top())
        self.emit(*line)
        self.indent()
        for k, v in node.__dict__.items():
            if v is None:
                continue
            self.push(k)
            self.visit(v)
            self.pop()

        self.dedent()

    def visit_list(self, L):
        if not self.empty():
            self.emit("%s:" % self.top())
        self.indent()
        for i, item in enumerate(L):
            self.push(i)
            self.visit(item)
            self.pop()
        self.dedent()

    def visit_object(self, obj):
        line = [obj]
        if not self.empty():
            line.insert(0, "%s:" % self.top())
        self.emit(*line)

    def emit(self, *bits):
        self.output.append(" " * self._level + " ".join(map(str, bits)))

    def empty(self):
        return not self._stack

    def push(self, item):
        self._stack.append(item)

    def pop(self):
        return self._stack.pop()

    def top(self):
        return self._stack[-1] if self._stack else None

    def indent(self):
        self._level += 2

    def dedent(self):
        self._level -= 2
        assert self._level >= 0, self._level



import pytest

from seldump.yaml import load_yaml
from seldump.config import get_config_errors


def test_empty_objs():
    """A config file with an empty list of objects is ok."""
    conf = load_yaml("db_objects: []")
    errors = get_config_errors(conf)
    assert not errors


@pytest.mark.parametrize("data", ["", "[]", "1", "42", '"hi"'])
def test_bad_start(data):
    """The config file must contain a dict."""
    conf = load_yaml(data)
    errors = get_config_errors(conf)
    assert len(errors) == 1
    assert "object" in errors[0]


@pytest.mark.parametrize("data", ["{}", "db_object: []", "{db_objects: [], foo: null}"])
def test_must_have_one_root(data):
    """The config entry must have precisely what expected."""
    conf = load_yaml(data)
    errors = get_config_errors(conf)
    assert errors


def test_bad_dbobjs():
    """The position of a bad entry in the main list is returned correctly."""
    conf = load_yaml(
        """\
db_objects:
  - null
  - "ouch"
  - true
  - 1
  - 1.1
  - []
"""
    )
    errors = get_config_errors(conf)
    assert len(errors) == 6
    for i, error in enumerate(errors):
        assert "<unicode string>:%s" % (i + 2) in error


def test_bad_items_pos():
    """The position of a bad entry in a dict/list is returned correctly."""
    conf = load_yaml(
        """\
db_objects:
  - name: 42
  - name: aaa
    kind: nope
  - name: bbb
    kinds:
     - table
     - boh
     - sequence
     - mah
"""
    )
    errors = get_config_errors(conf)
    assert len(errors) == 4
    assert "<unicode string>:2" in errors[0]
    assert "<unicode string>:4" in errors[1]
    assert "<unicode string>:8" in errors[2]
    assert "<unicode string>:10" in errors[3]


@pytest.mark.parametrize(
    "attr, value", (("name", "x"), ("schema", "y"), ("kind", "table"))
)
def test_no_singular_plural(attr, value):
    args = {"attr": attr, "value": value}
    conf = load_yaml(
        """\
db_objects:
  - %(attr)s: %(value)s
        """
        % args
    )
    errors = get_config_errors(conf)
    assert not errors

    conf = load_yaml(
        """\
db_objects:
  - %(attr)ss: [%(value)s]
        """
        % args
    )
    errors = get_config_errors(conf)
    assert not errors

    conf = load_yaml(
        """\
db_objects:
  - %(attr)s: %(value)s
    %(attr)ss: [%(value)s]
        """
        % args
    )
    errors = get_config_errors(conf)
    assert len(errors) == 1
    assert "string>:3" in errors[0]
    assert "can't specify both" in errors[0]


@pytest.mark.parametrize("attr", ("names", "schemas"))
def test_bad_regexpr(attr):
    conf = load_yaml(
        """\
db_objects:
  - %s: aaa
        """
        % attr
    )
    errors = get_config_errors(conf)
    assert not errors

    conf = load_yaml(
        """\
db_objects:
  - %s: aaa(
        """
        % attr
    )
    errors = get_config_errors(conf)
    assert len(errors) == 1
    assert "string>:2" in errors[0]
    assert "not a valid regular expression" in errors[0]


@pytest.mark.parametrize(
    "kind", ("table", "sequence", "partitioned table", "materialized view")
)
def test_valid_kind(kind):
    conf = load_yaml(
        """\
db_objects:
  - kind: %(kind)s
  - kinds: [%(kind)s]
        """
        % {"kind": kind}
    )
    errors = get_config_errors(conf)
    assert not errors


def test_bad_kind():
    conf = load_yaml(
        """\
db_objects:
  - kind: wat
  - kinds: [wat]
        """
    )
    errors = get_config_errors(conf)
    assert len(errors) == 2
    assert "string>:2" in errors[0]
    assert "'wat' is not one of" in errors[0]
    assert "string>:3" in errors[1]
    assert "'wat' is not one of" in errors[1]


def test_unexpected_attr():
    conf = load_yaml(
        """\
db_objects:
  - name: billy
    kid: table
        """
    )
    errors = get_config_errors(conf)
    assert len(errors) == 1
    assert "string>:3" in errors[0]
    assert "kid" in errors[0]
    assert "unexpected" in errors[0]



#!/usr/bin/env python3

"""
Reading object from a PostgreSQL database.

This file is part of pg_seldump.
"""

import logging
from functools import lru_cache

import psycopg
from psycopg import sql
from psycopg.rows import namedtuple_row

from .consts import DUMPABLE_KINDS, KIND_TABLE, KIND_PART_TABLE, REVKINDS
from .reader import Reader
from .dbobjects import DbObject, Column, ForeignKey
from .exceptions import DumpError

logger = logging.getLogger("seldump.dbreader")


class DbReader(Reader):
    def __init__(self, dsn):
        super().__init__()
        self.dsn = dsn

    @property
    @lru_cache(maxsize=1)
    def connection(self):
        logger.debug("connecting to '%s'", self.dsn)
        try:
            cnn = psycopg.connect(self.dsn, row_factory=namedtuple_row)
        except Exception as e:
            raise DumpError("error connecting to the database: %s" % e)

        cnn.autocommit = True
        return cnn

    def cursor(self):
        return self.connection.cursor()

    def obj_as_string(self, obj):
        """
        Convert a `psycopg.sql.Composable` object to string
        """
        return obj.as_string(self.connection)

    def load_schema(self):
        for rec in self._fetch_objects():
            obj = DbObject.from_kind(
                rec.kind,
                oid=rec.oid,
                schema=rec.schema,
                name=rec.name,
                extension=rec.extension,
                extcondition=rec.extcondition,
            )
            self.db.add_object(obj)

        for rec in self._fetch_columns():
            table = self.db.get(oid=rec.table_oid)
            assert table, "no table with oid %s for column %s found" % (
                rec.table_oid,
                rec.name,
            )
            col = Column(name=rec.name, type=rec.type)
            table.add_column(col)

        for rec in self._fetch_fkeys():
            table = self.db.get(oid=rec.table_oid)
            assert table, "no table with oid %s for foreign key %s found" % (
                rec.table_oid,
                rec.name,
            )
            ftable = self.db.get(oid=rec.ftable_oid)
            assert ftable, "no table with oid %s for foreign key %s found" % (
                rec.ftable_oid,
                rec.name,
            )
            fkey = ForeignKey(
                name=rec.name,
                table_oid=rec.table_oid,
                table_cols=rec.table_cols,
                ftable_oid=rec.ftable_oid,
                ftable_cols=rec.ftable_cols,
            )
            table.add_fkey(fkey)
            ftable.add_ref_fkey(fkey)

        for rec in self._fetch_sequences_deps():
            table = self.db.get(oid=rec.table_oid)
            assert table, "no table with oid %s for sequence %s found" % (
                rec.table_oid,
                rec.seq_oid,
            )
            seq = self.db.get(oid=rec.seq_oid)
            assert seq, "no sequence %s found" % rec.seq_oid
            self.db.add_sequence_user(seq, table, rec.column)

    def _fetch_objects(self):
        logger.debug("fetching database objects")
        with self.cursor() as cur:
            cur.execute(
                """
select
    r.oid as oid,
    s.nspname as schema,
    r.relname as name,
    r.relkind as kind,

    e.extname as extension,
    -- equivalent of
    -- extcondition[array_position(extconfig, r.oid)]
    -- but array_position not available < PG 9.5
    (
        select e.extcondition[row_number]
        from (
            select unnest, row_number() over ()
            from (select unnest(e.extconfig)) t0
        ) t1
        where unnest = r.oid
    ) as extcondition
from pg_class r
join pg_namespace s on s.oid = r.relnamespace
left join pg_extension ec on r.oid = any(ec.extconfig)
left join pg_extension e on ec.oid = e.oid
where r.relkind = any(%(stateless)s)
and s.nspname != 'information_schema'
and s.nspname !~ '^pg_'
order by s.nspname, r.relname
""",
                {"stateless": list(DUMPABLE_KINDS)},
            )
            return cur.fetchall()

    def _fetch_sequences_deps(self):
        logger.debug("fetching sequences dependencies")
        with self.cursor() as cur:
            cur.execute(
                """
select tbl.oid as table_oid, att.attname as column, seq.oid as seq_oid
from pg_depend dep
join pg_attrdef def
    on dep.classid = 'pg_attrdef'::regclass and dep.objid = def.oid
join pg_attribute att on (def.adrelid, def.adnum) = (att.attrelid, att.attnum)
join pg_class tbl on tbl.oid = att.attrelid
join pg_class seq
    on dep.refclassid = 'pg_class'::regclass
    and seq.oid = dep.refobjid
    and seq.relkind = 'S'
"""
            )
            return cur.fetchall()

    def _fetch_columns(self):
        logger.debug("fetching columns")
        with self.cursor() as cur:
            # attnum gives their order; attnum < 0 are system columns
            # attisdropped flags a dropped column.
            cur.execute(
                """
select
    attrelid as table_oid,
    attname as name,
    atttypid::regtype as type
from pg_attribute a
join pg_class r on r.oid = a.attrelid
join pg_namespace s on s.oid = r.relnamespace
where r.relkind = any(%(kinds)s)
and a.attnum > 0
and not attisdropped
and s.nspname != 'information_schema'
and s.nspname !~ '^pg_'
order by a.attrelid, a.attnum
                """,
                {"kinds": [REVKINDS[KIND_TABLE], REVKINDS[KIND_PART_TABLE]]},
            )
            return cur.fetchall()

    def _fetch_fkeys(self):
        logger.debug("fetching foreign keys")
        with self.cursor() as cur:
            cur.execute(
                """
select
    c.conname as name,
    c.conrelid as table_oid,
    array_agg(ra.attname) as table_cols,
    c.confrelid as ftable_oid,
    array_agg(fa.attname) as ftable_cols
from pg_constraint c
join (
    select oid, generate_series(1, array_length(conkey,1)) as attidx
    from pg_constraint
    where contype = 'f') exp on c.oid = exp.oid
join pg_attribute ra
    on (ra.attrelid, ra.attnum) = (c.conrelid, c.conkey[exp.attidx])
join pg_attribute fa
    on (fa.attrelid, fa.attnum) = (c.confrelid, c.confkey[exp.attidx])
join pg_class r on c.conrelid = r.oid
join pg_namespace rs on rs.oid = r.relnamespace
join pg_class fr on c.confrelid = fr.oid
join pg_namespace fs on fs.oid = fr.relnamespace
where rs.nspname != 'information_schema' and rs.nspname !~ '^pg_'
and   fs.nspname != 'information_schema' and fs.nspname !~ '^pg_'
group by 1, 2, 4
order by name
"""
            )
            return cur.fetchall()

    def get_sequence_value(self, seq):
        """
        Return the last value of a sequence.
        """
        with self.cursor() as cur:
            cur.execute(sql.SQL("select last_value from {}").format(seq.ident))
            val = cur.fetchone()[0]
            return val

    def copy(self, stmt, file):
        """
        Run a copy... to stdout statement.
        """
        with self.cursor() as cur:
            with cur.copy(stmt) as copy:
                for data in copy:
                    file.write(data)



from seldump.writer import Writer


class TestWriter(Writer):
    __test__ = False

    def __init__(self):
        self.reset()

    def reset(self):
        self.dump_started = False
        self.dump_ended = False
        self.closed = False
        self.dumped = []

    def begin_dump(self):
        assert not self.dump_started
        assert not self.dump_ended
        assert not self.closed
        self.dump_started = True

    def end_dump(self):
        assert self.dump_started
        assert not self.dump_ended
        assert not self.closed
        self.dump_ended = True

    def dump_table(self, table, match):
        assert self.dump_started
        assert not self.dump_ended
        assert not self.closed
        self.dumped.append((table, match))

    def dump_sequence(self, seq, match):
        assert self.dump_started
        assert not self.dump_ended
        assert not self.closed
        self.dumped.append((seq, match))

    def dump_materialized_view(self, matview, match):
        assert self.dump_started
        assert not self.dump_ended
        assert not self.closed
        self.dumped.append((matview, match))

    def close(self):
        assert self.dump_started == self.dump_ended
        self.closed = True



#!/usr/bin/env python3
"""
Pretend to write a dump

This file is part of pg_seldump.
"""

import logging

from .writer import Writer

logger = logging.getLogger("seldump.dummywriter")


class DummyWriter(Writer):
    def begin_dump(self):
        logger.debug("start of dump")

    def end_dump(self):
        logger.debug("end of dump")

    def dump_table(self, table, match):
        logger.info("would dump %s %s", table.kind, table)

    def dump_sequence(self, seq, match):
        logger.info("would dump %s %s", seq.kind, seq)

    def dump_materialized_view(self, matview, match):
        logger.info("would dump %s %s", matview.kind, matview)

    def close(self):
        pass



from seldump.dbobjects import Table


def test_fkey_nav(dumper, db):
    """
    Test one navigation step

        table1:dump -> table2:ref

        table3:unknown
    """
    dumper.reader.load_db(
        db.create_sample(3, fkeys=[("table1", "t2id", "table2", "id")])
    )
    dumper.add_config({"db_objects": [{"name": "table1"}]})
    dumper.perform_dump()
    objs = [obj for obj, match in dumper.writer.dumped if isinstance(obj, Table)]
    assert len(objs) == 2


def test_fkey_nav_rec(dumper, db):
    """
    Test navigation is transitive

        table1:dump -> table2:ref -> table3:ref

        table4:unknown
    """
    dumper.reader.load_db(
        db.create_sample(
            4,
            fkeys=[
                ("table1", "t2id", "table2", "id"),
                ("table2", "t3id", "table3", "id"),
            ],
        )
    )
    dumper.add_config({"db_objects": [{"name": "table1"}]})
    dumper.perform_dump()
    objs = [obj for obj, match in dumper.writer.dumped if isinstance(obj, Table)]
    assert len(objs) == 3


def test_fkey_nav_stops_on_skip(dumper, db):
    """
    Test navigation stops at a skip table

        table1:dump -> table2:ref -> table3:skip -> table4:unknown

        table5:unknown
    """
    dumper.reader.load_db(
        db.create_sample(
            5,
            fkeys=[
                ("table1", "t2id", "table2", "id"),
                ("table2", "t3id", "table3", "id"),
                ("table3", "t4id", "table4", "id"),
            ],
        )
    )
    dumper.add_config(
        {
            "db_objects": [
                {"name": "table1"},
                {"name": "table3", "action": "skip"},
            ]
        }
    )
    dumper.perform_dump()
    objs = [obj for obj, match in dumper.writer.dumped if isinstance(obj, Table)]
    assert len(objs) == 2


def test_two_referrers(dumper, db):
    """
    Test that table4 gets two referrers.

                    -> table2:ref ->
        table1:dump                  table4:ref
                    -> table3:ref ->

        table5:unknown
    """
    dumper.reader.load_db(
        db.create_sample(
            5,
            fkeys=[
                ("table1", "t2id", "table2", "id"),
                ("table1", "t3id", "table3", "id"),
                ("table2", "t24id", "table4", "id"),
                ("table3", "t34id", "table4", "id"),
            ],
        )
    )
    dumper.add_config({"db_objects": [{"name": "table1"}]})
    dumper.perform_dump()
    objs = [obj for obj, match in dumper.writer.dumped if isinstance(obj, Table)]
    assert len(objs) == 4

    (match,) = [match for obj, match in dumper.writer.dumped if obj.name == "table4"]
    assert len(match.referenced_by) == 2
    assert sorted(fkey.name for fkey in match.referenced_by) == [
        "t24id_table4_id_fkey",
        "t34id_table4_id_fkey",
    ]


def test_no_endless_loop(dumper, db):
    """
    Test we don't get stuck in a loop

        table1:dump -> table2:ref ->
                            ^        table3:ref
                       table4:ref <-
        table5:unknown
    """
    dumper.reader.load_db(
        db.create_sample(
            5,
            fkeys=[
                ("table1", "t12id", "table2", "id"),
                ("table2", "t3id", "table3", "id"),
                ("table3", "t4id", "table4", "id"),
                ("table4", "t42id", "table2", "id"),
            ],
        )
    )
    dumper.add_config({"db_objects": [{"name": "table1"}]})
    dumper.perform_dump()
    objs = [obj for obj, match in dumper.writer.dumped if isinstance(obj, Table)]
    assert len(objs) == 4

    (match,) = [match for obj, match in dumper.writer.dumped if obj.name == "table2"]
    assert len(match.referenced_by) == 2
    assert sorted(fkey.name for fkey in match.referenced_by) == [
        "t12id_table2_id_fkey",
        "t42id_table2_id_fkey",
    ]



from psycopg import sql

from seldump import query
from seldump.database import Database
from seldump.dbobjects import Table, Column, ForeignKey


def test_query_visit(fakeconn):
    """
    Transform the representation into the structure for the following query:

    select * from tab2 t0
    where exists (
        select 1 from (
            select t2id from tab1
            where {cond1}
        ) t1
        where t1.t2id = t0.id;
    )
    """
    db = Database()
    table1 = db.add_object(Table(oid=1, schema="public", name="table1"))
    table2 = db.add_object(Table(oid=2, schema="public", name="table2"))
    table1.add_column(Column("id", "integer"))
    table1.add_column(Column("t2id", "integer"))
    table2.add_column(Column("id", "integer"))
    table1.add_fkey(ForeignKey("t2id_fkey", 1, ["t2id"], 2, ["id"]))

    q = query.Select(
        from_=query.FromEntry(table2, alias="t0"),
        columns=[sql.SQL("*")],
        where=query.Exists(
            query.Select(
                from_=query.FromEntry(table1, alias="t1"),
                columns=[sql.SQL("1")],
                where=query.FkeyJoin(table1.get_fkey("t2id_fkey"), "t1", "t0"),
            )
        ),
    )

    v = query.SqlQueryVisitor()
    stmt = v.visit(q)
    assert stmt.as_string(fakeconn)



"""
Representation of database objects to handle by the program.

This file is part of pg_seldump.
"""


import re

from psycopg import sql

from . import consts


class DbObject:
    """
    An object in a database
    """

    __slots__ = (
        "oid",
        "schema",
        "name",
        "extension",
        "extcondition",
    )

    _kinds = {}

    kind = None

    def __init__(self, oid, schema, name, extension=None, extcondition=None):
        self.oid = oid
        self.schema = schema
        self.name = name
        self.extension = extension
        self.extcondition = extcondition

    @classmethod
    def from_kind(cls, kind, oid, schema, name, **kwargs):
        # Values from the database or from yaml
        if kind in consts.PG_KINDS:
            kind = consts.PG_KINDS[kind]
        if kind not in cls._kinds:
            raise ValueError("unknown db object kind: %s" % kind)
        return cls._kinds[kind](oid, schema, name, **kwargs)

    def __repr__(self):
        return "<%s %s at 0x%x>" % (self.__class__.__name__, self, id(self))

    def __str__(self):
        return self.escape_idents(self.schema, self.name)

    @property
    def ident(self):
        """The object name as an Identifier"""
        return sql.Identifier(self.schema, self.name)

    @classmethod
    def escape_idents(self, *args):
        """
        Convert tokens into dot-separated string to be safety merged to a query

        It gives a mmore readable but less secure version of ident, because it
        doesn't escape keywords; however it's a string so it can be just
        printed, without requiring a connection as Identifier.to_string() does.
        Use it only to display to humans, not to send to the database.
        """
        rv = []
        for arg in args:
            if re.match(r"^[a-z][a-z0-9_]*$", arg):
                rv.append(arg)
            else:
                arg = arg.replace('"', '""')
                rv.append('"%s"' % arg)

        return ".".join(rv)

    @staticmethod
    def register(kind):
        """
        Decorator to allow a class to be instantiated by `from_kind()`.
        """

        def register_(cls):
            if kind in DbObject._kinds:
                raise ValueError("the kind %s is already associated to a class" % kind)
            DbObject._kinds[kind] = cls
            cls.kind = kind
            return cls

        return register_


@DbObject.register(consts.KIND_TABLE)
class Table(DbObject):
    """A table in a database."""

    __slots__ = DbObject.__slots__ + (
        "columns",
        "_cols_by_name",
        "fkeys",
        "ref_fkeys",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.columns = []
        self._cols_by_name = {}
        self.fkeys = []
        self.ref_fkeys = []

    def add_column(self, column):
        if column.name in self._cols_by_name:
            raise ValueError(
                "the table %s has already a column called %s" % (self, column.name)
            )
        self.columns.append(column)
        self._cols_by_name[column.name] = column

    def get_column(self, name):
        return self._cols_by_name.get(name)

    def add_fkey(self, fkey):
        for col in fkey.table_cols:
            # using format because https://github.com/psf/black/issues/1259
            assert (
                col in self._cols_by_name
            ), "column {} in fkey {} is not in the table {}".format(col, fkey, self)
        self.fkeys.append(fkey)

    def get_fkey(self, name):
        for fkey in self.fkeys:
            if fkey.name == name:
                return fkey

    def add_ref_fkey(self, fkey):
        for col in fkey.ftable_cols:
            assert (
                col in self._cols_by_name
            ), "column {} in fkey {} is not in the table {}".format(col, fkey, self)
        self.ref_fkeys.append(fkey)


@DbObject.register(consts.KIND_PART_TABLE)
class PartitionedTable(Table):
    """A partitioned table in a database."""

    # TODO: pretty untested


@DbObject.register(consts.KIND_SEQUENCE)
class Sequence(DbObject):
    """A sequence table in a database."""


@DbObject.register(consts.KIND_MATVIEW)
class MaterializedView(DbObject):
    """A materialized view in a database."""


class Column:
    __slots__ = ("name", "type", "used_sequence_oids")

    def __init__(self, name, type):
        self.name = name
        self.type = type
        self.used_sequence_oids = []

    @property
    def ident(self):
        """The column name as an Identifier"""
        return sql.Identifier(self.name)

    def __repr__(self):
        return "<%s %s at 0x%x>" % (self.__class__.__name__, self, id(self))

    def __str__(self):
        return DbObject.escape_idents(self.name)

    def add_used_sequence(self, seq):
        if not seq.oid:
            raise ValueError("the sequence %s must have an oid" % seq)
        self.used_sequence_oids.append(seq.oid)


class ForeignKey:
    __slots__ = (
        "name",
        "table_oid",
        "table_cols",
        "ftable_oid",
        "ftable_cols",
    )

    def __init__(self, name, table_oid, table_cols, ftable_oid, ftable_cols):
        self.name = name
        self.table_oid = table_oid
        self.table_cols = table_cols
        self.ftable_oid = ftable_oid
        self.ftable_cols = ftable_cols

    @property
    def ident(self):
        """The fkey name as an Identifier"""
        return sql.Identifier(self.name)

    def __repr__(self):
        return "<%s %s at 0x%x>" % (
            self.__class__.__name__,
            self.name,
            id(self),
        )



#!/usr/bin/env python3
"""
Representation of a database to handle by the program.

This file is part of pg_seldump.
"""

from .dbobjects import Table


class Database:
    def __init__(self):
        self._objects = []
        self._by_oid = {}
        self._by_name = {}

    def clear(self):
        del self._objects
        self._by_oid.clear()
        self._by_name.clear()

    def add_object(self, obj):
        self._objects.append(obj)
        if obj.oid is not None:
            if obj.oid in self._by_oid:
                raise ValueError(
                    "the database already contains an object with oid %s" % obj.oid
                )
            self._by_oid[obj.oid] = obj

        key = (obj.schema, obj.name)
        if key in self._by_name:
            raise ValueError("the database already contains an object called %s" % obj)
        self._by_name[key] = obj

        return obj

    def get(self, schema=None, name=None, oid=None, cls=None):
        """
        Return an object from the database.

        The object can be specified by schema/name or by oid.
        It is possible to specify the class of the object espected.

        Return None if the object is not found or if it is not the right class.
        """
        if (schema is None) != (name is None):
            raise TypeError("you should either specify both schema and name or none")

        if schema is not None:
            rv = self._by_name.get((schema, name))
        elif oid:
            rv = self._by_oid.get(oid)
        else:
            raise TypeError("you should specify either schema/name or oid")

        # Return the object only if the right class
        if rv and cls is not None:
            if not isinstance(rv, cls):
                rv = None

        return rv

    def __iter__(self):
        yield from self._objects

    def get_tables_using_sequence(self, oid):
        rv = []
        for obj in self:
            if not isinstance(obj, Table):
                continue
            for col in obj.columns:
                if oid in col.used_sequence_oids:
                    rv.append((obj, col))

        return rv

    def add_sequence_user(self, seq, table, col_name):
        col = table.get_column(col_name)
        assert col, "no column %s in table %s" % (col_name, table)
        col.add_used_sequence(seq)



from collections import defaultdict

from seldump.reader import Reader
from seldump.dbobjects import DbObject, Column, Sequence, ForeignKey


class TestReader(Reader):
    __test__ = False

    def __init__(self):
        super().__init__()
        self.last_oid = 1

    def load_schema():
        pass

    def load_db(self, objs):
        used_seqs = defaultdict(list)

        for obj in objs:
            name = obj["name"]
            assert '"' not in name  # TODO: test with funny names
            assert "." not in name  # TODO: test with funny names

            schema = obj.get("schmea", "public")

            kind = obj.get("kind", "table")
            oid = obj.get("oid", None)
            if oid is None:
                oid = self.get_oid()

            dbobj = DbObject.from_kind(kind, oid, schema, name)
            if "columns" in obj:
                for colname, col in obj["columns"].items():
                    coltype = col["type"]
                    dbcol = Column(name=colname, type=coltype)
                    dbobj.add_column(dbcol)

                    if "use_sequence" in col:
                        used_seqs["public", col["use_sequence"]].append((dbobj, dbcol))

            self.db.add_object(dbobj)

        # create sequence if needed, attach them to columns
        for (schema, name), dbobjs in used_seqs.items():
            seq = self.db.get(schema, name, cls=Sequence)
            if seq is None:
                seq = Sequence(self.get_oid(), schema, name)
                self.db.add_object(seq)

            for obj, col in dbobjs:
                self.db.add_sequence_user(seq, obj, col.name)

        # create fkeys
        for obj in objs:
            if not obj.get("fkeys"):
                continue
            # TODO: other schemas
            t1 = self.db.get(obj.get("schema", "public"), obj["name"])
            assert t1
            for fkey in obj["fkeys"]:
                # TODO: other schemas
                t2 = self.db.get("public", fkey["ftable"])
                fkey = ForeignKey(
                    name=fkey["name"],
                    table_oid=t1.oid,
                    table_cols=fkey["cols"],
                    ftable_oid=t2.oid,
                    ftable_cols=fkey["fcols"],
                )
                t1.add_fkey(fkey)
                t2.add_ref_fkey(fkey)

    def get_oid(self):
        oid = self.last_oid
        self.last_oid += 1
        return oid



#!/usr/bin/env python3

"""
Program exceptions.

This file is part of pg_seldump.
"""


class SelDumpException(Exception):
    """A controlled exception raised by the script."""


class DumpError(SelDumpException):
    """Error dumping the database."""


class ConfigError(SelDumpException):
    """Error parsing configuration."""



#!/usr/bin/env python3
"""
Base classes to implement the visitor pattern.

https://en.wikipedia.org/wiki/Visitor_pattern

This file is part of pg_seldump.
"""


class Node:
    """
    A node which can accept a NodeVisitor.

    This impleentation is Liskov-friendly.
    """


class NodeVisitor:
    def visit(self, node):
        for cls in node.__class__.__mro__:
            meth = getattr(self, "visit_" + cls.__name__, None)
            if meth is not None:
                return meth(node)

        else:
            # We expect at list visit_object to be implemented
            assert False, "no visitor method found in %r" % self

    def visit_object(self, node):
        raise NotImplementedError(
            "visitor %s cannot handle node %s"
            % (self.__class__.__name__, node.__class__.__name__)
        )



#!/usr/bin/env python3

"""
Readers base class

This file is part of pg_seldump.
"""

from abc import ABC, abstractmethod


class Reader(ABC):
    """
    The base class of an object to read e db to dump.
    """

    def __init__(self):
        self.db = None

    @abstractmethod
    def load_schema(self):
        pass



#!/usr/bin/python3

from seldump import query


def test_subqueries(dumper, db, fakeconn):
    """
    Test one navigation step

        table1:dump -> table2:ref -> table3:ref

    """
    dumper.reader.load_db(
        db.create_sample(
            4,
            fkeys=[
                ("table1", "t2id", "table2", "id"),
                ("table2", "t3id", "table3", "id"),
            ],
        )
    )
    dumper.add_config({"db_objects": [{"name": "table1", "filter": "data='aaa'"}]})
    dumper.plan_dump()

    q = dumper.matches[dumper.db.get("public", "table3").oid].query
    assert isinstance(q, query.Select)
    assert q.from_.source.name == "table3"
    assert isinstance(q.where, query.Exists)

    sq = q.where.query
    assert sq.from_.source.name == "table2"
    assert isinstance(sq.where, query.And)
    assert sq.where.conds[0].fkey.name == "t3id_table3_id_fkey"
    assert isinstance(sq.where.conds[1], query.Exists)

    sq = sq.where.conds[1].query
    assert sq.from_.source.name == "table1"
    assert isinstance(sq.where, query.And)
    assert sq.where.conds[0].fkey.name == "t2id_table2_id_fkey"
    assert sq.where.conds[1]._obj == "data='aaa'"

    stmt = query.SqlQueryVisitor().visit(q)
    assert stmt.as_string(fakeconn)


def test_two_dependencies(db, dumper, fakeconn):
    """
    A fkey table can be referred by more than one table.

        table1:dump ->
                        table3:ref -> table4:ref
        table2:dump ->
    """

    objs = db.create_sample(
        4,
        fkeys=[
            ("table1", "t13id", "table3", "id"),
            ("table2", "t23id", "table3", "id"),
            ("table3", "t4id", "table4", "id"),
        ],
    )
    dumper.reader.load_db(objs)

    dumper.add_config(
        """
db_objects:
- name: table1
  filter: data <= 'b'

- name: table2
  filter: data <= 'f'
"""
    )
    dumper.plan_dump()

    q = dumper.matches[dumper.db.get("public", "table4").oid].query
    assert isinstance(q, query.Select)
    assert q.from_.source.name == "table4"
    assert isinstance(q.where, query.Exists)

    sq = q.where.query
    assert sq.from_.source.name == "table3"
    assert isinstance(sq.where, query.And)
    assert sq.where.conds[0].fkey.name == "t4id_table4_id_fkey"
    assert isinstance(sq.where.conds[1], query.Or)
    assert list(map(type, (sq.where.conds[1].conds))) == [query.Exists] * 2

    sqs = [cond.query for cond in sq.where.conds[1].conds]
    sqs.sort(key=lambda sq: sq.from_.source.name)
    assert [sq.from_.source.name for sq in sqs] == ["table1", "table2"]
    assert [sq.where.conds[0].fkey.name for sq in sqs] == [
        "t13id_table3_id_fkey",
        "t23id_table3_id_fkey",
    ]
    assert [sq.where.conds[1]._obj for sq in sqs] == [
        "data <= 'b'",
        "data <= 'f'",
    ]



#!/usr/bin/env python3
"""
Writers base class

This file is part of pg_seldump.
"""

from abc import ABC, abstractmethod


class Writer(ABC):
    """
    The base class of an object to write a db dump.
    """

    @abstractmethod
    def begin_dump(self):
        pass

    @abstractmethod
    def end_dump(self):
        pass

    @abstractmethod
    def dump_table(self, table, match):
        pass

    @abstractmethod
    def dump_sequence(self, seq, match):
        pass

    @abstractmethod
    def dump_materialized_view(self, matview, match):
        pass



from seldump.dbobjects import Sequence, Table


def test_read_something(dbdumper, db):
    objs = db.create_sample(2, fkeys=[("table1", "t2id", "table2", "id")])
    db.write_schema(objs)
    dbdumper.reader.load_schema()

    ts = {}
    for name in "table1", "table2":
        ts[name] = table = dbdumper.db.get("public", name, cls=Table)
        assert table
        assert table.oid > 1024
        assert table.name == name
        assert table.get_column("id")
        assert table.get_column("data")

        seq = dbdumper.db.get("public", name + "_id_seq", cls=Sequence)
        assert seq

        ((t, c),) = dbdumper.db.get_tables_using_sequence(seq.oid)
        assert t is table
        assert c is table.get_column("id")

    assert ts["table1"].get_column("t2id")

    (fk,) = ts["table1"].fkeys
    assert fk.table_oid == ts["table1"].oid
    assert fk.table_cols == ["t2id"]
    assert fk.ftable_oid == ts["table2"].oid
    assert fk.ftable_cols == ["id"]



import pytest

from seldump.dbobjects import Table, Sequence


@pytest.mark.parametrize(
    "details, dumped",
    [
        ({}, True),
        ({"no_columns": ["data"]}, True),
        ({"no_columns": ["id"]}, False),
        ({"replace": {"data": "NULL"}}, True),
        ({"replace": {"id": "NULL"}}, False),
    ],
)
def test_sequence_skipped(dumper, db, details, dumped):
    dumper.reader.load_db(db.create_sample(2))

    tbl = dumper.db.get("public", "table1")
    assert isinstance(tbl, Table)
    seq = dumper.db.get("public", "table1_id_seq")
    assert isinstance(seq, Sequence)

    conf = {"name": "table1"}
    conf.update(details)
    dumper.add_config({"db_objects": [conf]})
    dumper.perform_dump()
    objs = [obj for obj, rule in dumper.writer.dumped]
    assert tbl in objs
    if dumped:
        assert len(objs) == 2
        assert seq in objs
    else:
        assert len(objs) == 1
        assert seq not in objs


def test_sequence_skip_override(dumper, db):
    dumper.reader.load_db(db.create_sample(2))
    tbl = dumper.db.get("public", "table1")
    seq = dumper.db.get("public", "table1_id_seq")
    dumper.add_config(
        {
            "db_objects": [
                {"name": "table1"},
                {"kind": "sequence", "action": "skip"},
            ]
        }
    )

    dumper.perform_dump()
    objs = [obj for obj, rule in dumper.writer.dumped]
    assert len(objs) == 1
    assert seq not in objs
    assert tbl in objs



#!/usr/bin/env python3
"""
pg_seldump -- setup script
"""

# Copyright (C) 2020 Daniele Varrazzo

# This file is part of pg_seldump


import re
import os
from setuptools import setup

# Grab the version without importing the module
# or we will get import errors on install if prerequisites are still missing
fn = os.path.join(os.path.dirname(__file__), "seldump/consts.py")
with open(fn) as f:
    m = re.search(r"""(?mi)^VERSION\s*=\s*["']+([^'"]+)["']+""", f.read())
if m:
    version = m.group(1)
else:
    raise ValueError("cannot find VERSION in the consts module")

# Read the description from the README
with open("README.rst") as f:
    readme = f.read()

classifiers = """
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Developers
Intended Audience :: Information Technology
Intended Audience :: Science/Research
Intended Audience :: System Administrators
License :: OSI Approved :: BSD License
Operating System :: POSIX
Programming Language :: Python :: 3
Topic :: Database
Topic :: Software Development
Topic :: Software Development :: Testing
Topic :: System :: Archiving :: Backup
Topic :: System :: Systems Administration
Topic :: Utilities
"""

requirements = """
psycopg >= 3.1
jsonschema >= 4.17
PyYAML >= 6
"""

setup(
    name="pg_seldump",
    description=readme.splitlines()[0],
    long_description="\n".join(readme.splitlines()[2:]).lstrip(),
    author="Daniele Varrazzo",
    author_email="daniele.varrazzo@gmail.com",
    url="https://github.com/dvarrazzo/pg_seldump",
    license="BSD",
    python_requires=">= 3.7",
    install_requires=requirements,
    extras_require={
        "test": [
            "pytest >= 7.2",
        ],
        "dev": [
            "black",
            "flake8",
        ],
    },
    packages=["seldump"],
    package_data={"seldump": ["schema/*.yaml"]},
    entry_points={"console_scripts": ["pg_seldump = seldump.cli:script"]},
    classifiers=[x for x in classifiers.split("\n") if x],
    zip_safe=False,
    version=version,
)



#!/usr/bin/env python3

"""
Object to perform a database dump.

This file is part of pg_seldump.
"""

import re
import logging
from operator import attrgetter

from psycopg import sql

from . import query
from .config import load_yaml, get_config_errors
from .database import Database
from .dumprule import DumpRule, RuleMatch
from .dbobjects import MaterializedView, Sequence, Table
from .exceptions import ConfigError, DumpError

logger = logging.getLogger("seldump.dumper")


class Dumper:
    """
    The logic of a database dump.
    """

    def __init__(self, reader, writer):
        self.db = Database()
        self.reader = reader
        self.writer = writer
        self.rules = []
        self.matches = {}

    @property
    def reader(self):
        return self._reader

    @reader.setter
    def reader(self, reader):
        self._reader = reader
        reader.db = self.db

    def clear(self):
        self.db.clear()
        del self.rules[:]
        self.matches.clear()

    def add_config(self, cfg):
        """
        Add a new config structure to the dumper

        The structure is what parsed by a json file. It must have a list
        of rules called 'db_objects'.

        You can parse a string too, which will be parsed from json.
        """
        if isinstance(cfg, str):
            # This case is mostly used for testing, so not really caring about
            # returning all the errors.
            cfg = load_yaml(cfg)
            errors = get_config_errors(cfg)
            if errors:
                raise ConfigError(errors[0])

        for cfg in cfg["db_objects"]:
            self.rules.append(DumpRule.from_config(cfg))

    def perform_dump(self):
        """
        Perform the dump of a database.

        Read schema and data from the reader, apply the configured rule,
        use the writer to emit dump data.
        """
        self.plan_dump()
        self.run_dump()

    def plan_dump(self):
        """
        Read config, source db, and calculate the operations to perform.

        This step doesn't need a writer.
        """
        self.find_matches()
        self.generate_statements()
        self.report_errors()

    def run_dump(self):
        """
        Perform a dump running the steps previously planned.
        """
        if self.writer is None:
            raise ValueError("no writer set")
        self.apply_actions()

    def find_matches(self):
        """
        Scan the object in the database and create the matches to dump them.
        """
        self.matches.clear()

        # Associate a match to every object of the database
        for obj in self.db:
            assert obj.oid, "by now, every object should have an oid"
            assert obj.oid not in self.matches, "oid {} is duplicate".format(obj.oid)
            match = self.get_match(obj)
            self.matches[obj.oid] = match

        # Find tables we need data to fulfill fkeys
        for obj in self.db:
            if not isinstance(obj, Table):
                continue
            if self.matches[obj.oid].action not in (
                DumpRule.ACTION_DUMP,
                DumpRule.ACTION_REFERENCED,
            ):
                continue
            self._add_referred_tables(obj)

        # Find unmentioned sequences and check if any table depend on them
        for obj in self.db:
            if not isinstance(obj, Sequence):
                continue
            if self.matches[obj.oid].action != DumpRule.ACTION_UNKNOWN:
                continue
            match = self._get_sequence_dependency_match(obj)
            if match is not None:
                self.matches[obj.oid] = match

    def generate_statements(self):
        gen = StatementsGenerator(self)
        for obj in self.db:
            if not isinstance(obj, Table):
                continue
            match = self.matches[obj.oid]
            gen.make_statements(obj, match)

    def report_errors(self):
        """
        If any error has been found report it and terminate the operations.
        """
        # search and report errors
        has_errors = False
        for obj in self.db:
            match = self.matches[obj.oid]
            if match.errors:
                has_errors = True
                for error in match.errors:
                    logger.error("cannot dump %s: %s", obj, error)

        if has_errors:
            raise DumpError()

    def apply_actions(self):
        """
        Apply all the requested actions.
        """
        # Dump sequences after tables: in case a sequence is pushed forward
        # by a operation it's better to have a hole in the sequence than to
        # have it returning a value already consumed.
        # Refresh the materialized views at the end.
        def key(obj):
            if isinstance(obj, Table):
                return 1
            elif isinstance(obj, Sequence):
                return 2
            # TODO: actually matviews should be dumped in dependency order.
            elif isinstance(obj, MaterializedView):
                return 3
            assert False, "what's the position of %s?" % obj

        objs = list(self.db)
        objs.sort(key=key)

        self.writer.begin_dump()

        for obj in objs:
            match = self.matches[obj.oid]
            meth = getattr(self, "_apply_" + match.action, None)
            if meth is None:
                raise DumpError("cannot dump a match %s", match.action)

            meth(obj, match)

        self.writer.end_dump()

    def _add_referred_tables(self, table, seen=None):
        logger.debug("exploring %s foreign keys", table)
        if seen is None:
            seen = set()

        if table.oid in seen:
            return

        seen.add(table.oid)

        for fkey in table.fkeys:
            logger.debug("found fkey %s", fkey.name)
            fmatch = self.matches[fkey.ftable_oid]
            if fmatch.action not in (
                DumpRule.ACTION_UNKNOWN,
                DumpRule.ACTION_REFERENCED,
                DumpRule.ACTION_DUMP,
            ):
                # skip, dump, error: we don't have to navigate it tho
                continue

            if fmatch.action == DumpRule.ACTION_UNKNOWN:
                fmatch.action = DumpRule.ACTION_REFERENCED
            if fkey not in fmatch.referenced_by:
                fmatch.referenced_by.append(fkey)

            self._add_referred_tables(fmatch.obj, seen)

    def _get_sequence_dependency_match(self, seq):
        for table, column in self.db.get_tables_using_sequence(seq.oid):
            ta = self.matches[table.oid]
            if ta.action not in (
                DumpRule.ACTION_DUMP,
                DumpRule.ACTION_REFERENCED,
            ):
                continue

            if column.name in ta.no_columns:
                logger.debug(
                    "%s %s depends on %s.%s which is not dumped",
                    seq.kind,
                    seq,
                    table,
                    column.name,
                )
                continue

            if column.name in ta.replace:
                logger.debug(
                    "%s %s depends on %s.%s which is replaced",
                    seq.kind,
                    seq,
                    table,
                    column.name,
                )
                continue

            # we found a table wanting this sequence
            logger.debug(
                "%s %s is needed by matched %s %s",
                seq.kind,
                seq,
                table.kind,
                table,
            )
            match = RuleMatch(seq, action=DumpRule.ACTION_REFERENCED)
            return match

    def get_match(self, obj):
        """
        Return the best matching rule for an object, None if none found
        """
        if obj.extension is not None and obj.extcondition is None:
            logger.debug(
                "%s %s in extension %s has no dump condition: skipping",
                obj.kind,
                obj,
                obj.extension,
            )
            return RuleMatch(obj, action=DumpRule.ACTION_SKIP)

        rules = [rule for rule in self.rules if rule.match(obj)]
        if not rules:
            return RuleMatch(obj, action=DumpRule.ACTION_UNKNOWN)

        rules.sort(key=attrgetter("score"), reverse=True)
        if len(rules) > 1 and rules[0].score == rules[1].score:
            raise ConfigError(
                "%s %s matches more than one rule: at %s and %s"
                % (obj.kind, obj, rules[0].pos, rules[1].pos)
            )

        return RuleMatch.from_rule(obj, rules[0])

    #
    # Methods to apply rules after statements have been generated
    # (dynamic dispatch from `apply_actions()`)
    #

    def _apply_unknown(self, obj, match):
        logger.debug("%s %s doesn't match any rule: skipping", obj.kind, obj)

    def _apply_skip(self, obj, match):
        logger.debug("skipping %s %s", obj.kind, obj)

    def _apply_dump(self, obj, match):
        meth = getattr(self.writer, "dump_" + obj.kind.replace(" ", "_"), None)
        if meth is None:
            raise DumpError("don't know how to dump objects of kind %s" % obj.kind)
        meth(obj, match)

    def _apply_ref(self, obj, match):
        self._apply_dump(obj, match)


class StatementsGenerator:
    """
    An object which can generate SQL statements out of the dump state.

    SQL statements are first generates into an intermediate structure (as
    seldump.query objects), then converted into a query (as psycopg.sql
    objects) thanks to the SqlQueryVisitor.
    """

    def __init__(self, dumper):
        self.dumper = dumper
        self.db = self.dumper.db
        self._alias_seq = 0

    def make_statements(self, table, match):
        """
        Set the statements to be used by the dump operation on `match`.
        """
        if match.action not in (
            DumpRule.ACTION_DUMP,
            DumpRule.ACTION_REFERENCED,
        ):
            return

        # Discard quietly a table with no column
        if not table.columns:
            match.action = DumpRule.ACTION_SKIP
            return

        self.find_errors(table, match)
        if match.errors:
            return

        self.set_copy_statement(table, match)
        self.set_import_statement(table, match)

    def find_errors(self, table, match):
        """
        Verify correctness of the operation and set errors on `match`.
        """
        for col in match.no_columns:
            if table.get_column(col) is None:
                match.errors.append(
                    "the table doesn't have the column '%s'"
                    " specified in 'no_columns'" % col
                )
        for col in match.replace:
            if table.get_column(col) is None:
                match.errors.append(
                    "the table doesn't have the column '%s'"
                    " specified in 'replace'" % col
                )

        if len(set(match.no_columns)) == len(table.columns):
            match.errors.append(
                "the table has no column left to dump: you should skip it"
            )

    def set_import_statement(self, table, match):
        """
        Set the statement used to import back data on the `match`.

        This statement will usually be printed on output, and will be a COPY
        FROM STDIN to read data from the rest of the file.
        """
        attrs = [col.ident for col in table.columns if col.name not in match.no_columns]

        match.import_statement = sql.SQL("\ncopy {} ({}) from stdin;\n").format(
            table.ident, sql.SQL(", ").join(attrs)
        )

    def set_copy_statement(self, table, match):
        """
        Set the statement used to extract data from the db on the `match`.

        This statement will be a COPY TO STDOUT that will be executed ad dump
        time and the output will be added to the dump file.

        The function also sets `match.query`, the query generated.
        """
        # If False can use "copy table (attrs) to stdout" to dump data.
        # Otherwise must use a slower "copy (query) to stdout"
        if not (
            match.action != DumpRule.ACTION_DUMP
            or match.replace
            or match.filter
            or table.extcondition
            or table.ref_fkeys
        ):
            self._set_copy_to_simple(table, match)
        else:
            match.query = self.make_query(table, match)
            copy = query.CopyOut(match.query)
            match.copy_statement = query.SqlQueryVisitor().visit(copy)

    def _set_copy_to_simple(self, table, match):
        attrs = self._get_dump_attrs(table, match)
        match.copy_statement = sql.SQL("copy {} ({}) to stdout").format(
            table.ident, sql.SQL(", ").join(attrs)
        )

    def make_query(self, table, match):
        """
        Generate the query to execute the desired `match`.
        """
        self._alias_seq = 0

        select = self._get_select(table, match)
        select.columns = self._get_dump_attrs(table, match)
        return select

    def _get_existence(self, table, fkey, parent, seen):
        """
        Return the Exists predicate to limit the query to a fkey content.
        """
        assert fkey.ftable_oid == table.oid
        ptable = self.db.get(oid=fkey.table_oid)
        pmatch = self.dumper.matches[fkey.table_oid]
        subsel = self._get_select(ptable, pmatch, seen=seen)
        fkj = query.FkeyJoin(fkey=fkey, from_=subsel.from_.alias, to=parent)

        return query.Exists(
            query=query.Select(
                columns=[sql.SQL("1")],
                from_=subsel.from_,
                where=self._maybe_and([fkj, subsel.where]),
            )
        )

    def _get_select(self, table, match, seen=None):
        """
        Return the part of the select common to subqueries and external ones.

        The returned select has the attribute list empty: the caller will have
        to populate it.
        """
        alias = self._get_alias()
        where = [self._get_filters(table, match)]
        seen = (seen or set()) | {table.oid}

        srfkeys = []
        for fkey in match.referenced_by:
            # self-referential fkeys are dealt with later
            if fkey.table_oid == fkey.ftable_oid:
                srfkeys.append(fkey)
                continue
            elif fkey.table_oid in seen:
                logger.warning("not going recursive for now")
                continue
            where.append(self._get_existence(table, fkey, parent=alias, seen=seen))

        q = query.Select(
            columns=[],
            from_=query.FromEntry(table, alias=alias),
            where=self._maybe_or(where),
        )

        # if there are self-referential fkeys q is the base of a recursive cte
        if srfkeys:
            rec_alias = alias + "r"
            rec_cond = self._maybe_or(
                [query.FkeyJoin(fkey, rec_alias, alias) for fkey in srfkeys]
            )

            q.columns = [sql.SQL("{}.*").format(sql.Identifier(alias))]
            q = query.Select(
                columns=[],
                from_=query.FromEntry(
                    query.RecursiveCTE(name=rec_alias, base_query=q, rec_cond=rec_cond),
                    alias=rec_alias,
                ),
            )

        return q

    def _get_filters(self, table, match):
        rv = []
        if table.extcondition:
            rv.append(sql.SQL(re.replace(r"(?i)^\s*where\s+", table.extcondition, "")))
        if match.filter:
            rv.append(sql.SQL(match.filter.strip()))

        return self._maybe_and(rv)

    def _get_dump_attrs(self, table, match):
        rv = []
        for col in table.columns:
            if col.name in match.no_columns:
                continue

            if col.name in match.replace:
                rv.append(
                    sql.SQL("({})").format(sql.SQL(match.replace[col.name].strip()))
                )
            else:
                rv.append(col.ident)

        return rv

    def _maybe_and(self, conds):
        return self._maybe_op(conds, query.And)

    def _maybe_or(self, conds):
        return self._maybe_op(conds, query.Or)

    def _maybe_op(self, conds, op):
        conds = [c for c in conds if c is not None]
        if not conds:
            return None
        elif len(conds) == 1:
            return conds[0]
        else:
            return op(conds)

    def _get_alias(self):
        rv = "t%s" % self._alias_seq
        self._alias_seq += 1
        return rv



#!/usr/bin/env python3
"""
Customized YAML parser.

This file is part of pg_seldump.
"""

import yaml

import logging

logger = logging.getLogger("seldump.yaml")


class ListWithPos(list):
    """
    A list with attached filename and line numbers where it was parsed from
    """

    __slots__ = ("filename", "lineno", "itemlines")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = None
        self.lineno = None
        self.itemlines = []

    def __repr__(self):
        return super().__repr__()

    def extend(self, other):
        for item in other:
            if isinstance(item, ScalarWithPos):
                self.itemlines.append(item.lineno)
                self.append(item.obj)
            else:
                self.itemlines.append(None)
                self.append(item)


class DictWithPos(dict):
    """
    A dict with attached filename and line numbers where it was parsed from
    """

    __slots__ = ("filename", "lineno", "itemlines")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = None
        self.lineno = None
        self.itemlines = {}

    def __repr__(self):
        return super().__repr__()

    def update(self, other):
        assert isinstance(other, dict)
        for k, v in other.items():
            if isinstance(k, ScalarWithPos):
                self.itemlines[k.obj] = k.lineno
                k = k.obj

            if isinstance(v, ScalarWithPos):
                # The value may be more precise than the key
                self.itemlines[k] = v.lineno
                v = v.obj

            self[k] = v


class ScalarWithPos:
    __slots__ = ("obj", "lineno")

    def __init__(self, obj, node):
        self.obj = obj
        self.lineno = node.start_mark.line + 1

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.obj, self.lineno)

    def __hash__(self):
        return hash(self.obj)

    def __getattr__(self, attr):
        return getattr(self.obj, attr)


class RichLoader(yaml.SafeLoader):
    """
    YAML parser storing file name and line number for each parsed dicts.
    """

    def construct_scalar(self, node):
        rv = super().construct_scalar(node)
        return ScalarWithPos(rv, node)

    def construct_yaml_null(self, node):
        rv = super().construct_yaml_null(node)
        return ScalarWithPos(rv, node)

    def construct_yaml_bool(self, node):
        rv = super().construct_yaml_bool(node)
        return ScalarWithPos(rv, node)

    def construct_yaml_int(self, node):
        rv = super().construct_yaml_int(node)
        return ScalarWithPos(rv, node)

    def construct_yaml_float(self, node):
        rv = super().construct_yaml_float(node)
        return ScalarWithPos(rv, node)

    def construct_yaml_map(self, node):
        data = DictWithPos()
        data.filename = node.start_mark.name
        data.lineno = node.start_mark.line + 1
        yield data
        value = self.construct_mapping(node)
        data.update(value)

    def construct_yaml_seq(self, node):
        data = ListWithPos()
        data.filename = node.start_mark.name
        data.lineno = node.start_mark.line + 1
        yield data
        data.extend(self.construct_sequence(node))


# This really overrides the yaml parser
RichLoader.add_constructor("tag:yaml.org,2002:null", RichLoader.construct_yaml_null)
RichLoader.add_constructor("tag:yaml.org,2002:bool", RichLoader.construct_yaml_bool)
RichLoader.add_constructor("tag:yaml.org,2002:int", RichLoader.construct_yaml_int)
RichLoader.add_constructor("tag:yaml.org,2002:float", RichLoader.construct_yaml_float)
RichLoader.add_constructor("tag:yaml.org,2002:map", RichLoader.construct_yaml_map)
RichLoader.add_constructor("tag:yaml.org,2002:seq", RichLoader.construct_yaml_seq)


def load_yaml(stream):
    """Load a yaml file from a file."""
    return yaml.load(stream, Loader=RichLoader)



#!/usr/bin/env python3

"""
Rule to configure dumping of certain db objects.

This file is part of pg_seldump.
"""

import re
import logging
from functools import lru_cache

logger = logging.getLogger("seldump.dumprule")


class DumpRule:
    """
    Dump configuration of a set of database objects

    Each DumpRule has a few selector attributes, to choose which objects it
    applies, and a set of attributes specifying what action to take.
    """

    ACTION_DUMP = "dump"
    ACTION_SKIP = "skip"
    ACTION_ERROR = "error"
    ACTION_REFERENCED = "ref"
    ACTION_UNKNOWN = "unknown"

    def __init__(
        self,
        name=None,
        names=None,
        schema=None,
        schemas=None,
        kind=None,
        kinds=None,
        adjust_score=0,
        action=ACTION_DUMP,
        no_columns=None,
        replace=None,
        filter=None,
    ):
        """
        The constructor arguments are the attributes from a valid rule object
        validated according the `schema/config.yaml` rules
        """
        # Matching attributes
        if name is not None:
            names = [name]

        if names is None:
            self.names = set()
            self.names_re = None
        elif isinstance(names, list):
            self.names = set(names)
            self.names_re = None
        else:
            self.names = set()
            self.names_re = re.compile(names, re.VERBOSE)

        if schema is not None:
            schemas = [schema]

        if schemas is None:
            self.schemas = set()
            self.schemas_re = None
        elif isinstance(schemas, list):
            self.schemas = set(schemas)
            self.schemas_re = None
        else:
            self.schemas = set()
            self.schemas_re = re.compile(schemas, re.VERBOSE)

        if kind is not None:
            kinds = [kind]
        self.kinds = set(kinds or ())

        self.adjust_score = adjust_score

        # Actions
        self.action = action
        self.no_columns = no_columns or []
        self.replace = replace or {}
        self.filter = filter

        # Description
        self.filename = None
        self.lineno = None

    @classmethod
    def from_config(cls, cfg):
        """
        Create a rule from a config object.

        It will also try to store the config position if found.
        """
        rv = cls(**cfg)
        rv.filename = getattr(cfg, "filename", None)
        rv.lineno = getattr(cfg, "lineno", None)
        return rv

    @property
    @lru_cache(maxsize=1)
    def score(self):
        """
        The score of the rule: the higher the stronger
        """
        score = self.adjust_score
        if self.names:
            score += 1000
        if self.names_re:
            score += 500
        if self.schemas:
            score += 100
        if self.schemas_re:
            score += 50
        if self.kinds:
            score += 10
        return score

    @property
    def pos(self):
        """
        Return the file name and line no where the rule was parsed.
        """
        return "%s:%s" % (self.filename, self.lineno)

    def match(self, obj):
        """
        Return True if the db object *obj* matches the rule.
        """
        if self.names and obj.name not in self.names:
            return False

        if self.names_re is not None and not self.names_re.match(obj.name):
            return False

        if self.schemas and obj.schema not in self.schemas:
            return False

        if self.schemas_re is not None and not self.schemas_re.match(obj.schema):
            return False

        if self.kinds and obj.kind not in self.kinds:
            return False

        return True


class RuleMatch:
    """
    The match between an Rule iand a database object.

    The object is closely connected to a DumpRule, but it extends on that,
    defining an operation taking into account the state of the database too.
    """

    def __init__(self, obj, action=DumpRule.ACTION_UNKNOWN):
        self.obj = obj
        self.action = action

        self.rule = None
        self.no_columns = []
        self.replace = {}
        self.filter = None

        self.referenced_by = []

        self.query = None
        self.import_statement = None
        self.copy_statement = None
        self.errors = []

    @classmethod
    def from_rule(cls, obj, rule):
        rv = cls(obj, rule.action)
        rv.action = rule.action
        rv.no_columns = rule.no_columns
        rv.replace = rule.replace
        rv.filter = rule.filter
        if rule.action == DumpRule.ACTION_ERROR:
            if rule.filename and rule.lineno:
                msg = "the object matches the error rule at %s:%s" % (
                    rule.filename,
                    rule.lineno,
                )
            else:
                msg = "the object matches an error rule"

            rv.errors.append(msg)

        return rv
